/**
* Copyright 2012 University of Massachusetts Amherst
* 
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
* 
*   http://www.apache.org/licenses/LICENSE-2.0
*   
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*/
package com.googlecode.clearnlp.bin;

import java.io.File;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.TimeUnit;

import org.kohsuke.args4j.CmdLineException;
import org.kohsuke.args4j.CmdLineParser;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import com.carrotsearch.hppc.IntDoubleMap;
import com.carrotsearch.hppc.IntObjectOpenHashMap;
import com.carrotsearch.hppc.ObjectIntOpenHashMap;
import com.googlecode.clearnlp.classification.algorithm.AdaGrad;
import com.googlecode.clearnlp.classification.model.AbstractModel;
import com.googlecode.clearnlp.classification.train.AbstractTrainSpace;
import com.googlecode.clearnlp.classification.train.StringTrainSpace;
import com.googlecode.clearnlp.component.AbstractComponent;
import com.googlecode.clearnlp.engine.EngineGetter;
import com.googlecode.clearnlp.io.FileExtFilter;
import com.googlecode.clearnlp.reader.AbstractColumnReader;
import com.googlecode.clearnlp.reader.AbstractReader;
import com.googlecode.clearnlp.reader.JointReader;
import com.googlecode.clearnlp.reader.LineReader;
import com.googlecode.clearnlp.reader.RawReader;
import com.googlecode.clearnlp.run.LiblinearTrain;
import com.googlecode.clearnlp.util.UTXml;
import com.googlecode.clearnlp.util.pair.Pair;

/**
 * @since 1.3.0
 * @author Jinho D. Choi ({@code jdchoi77@gmail.com})
 */
abstract public class AbstractBin
{
	final public String TAG_READER					= "reader";
	final public String TAG_TYPE					= "type";
	final public String TAG_READER_COLUMN			= "column";
	final public String TAG_READER_COLUMN_INDEX		= "index";
	final public String TAG_READER_COLUMN_FIELD		= "field";
	
	final public String TAG_TRAIN					= "train";
	final public String TAG_TRAIN_ALGORITHM			= "algorithm";
	final public String TAG_TRAIN_ALGORITHM_NAME	= "name";
	final public String TAG_TRAIN_THREADS			= "threads";
	
	final public String TAG_LANGUAGE				= "language";
	final public String TAG_MODE					= "mode";
	final public String TAG_MODEL					= "model";
	final public String TAG_MODELS					= "models";
	final public String TAG_PATH					= "path";
	
	/** Initializes arguments using args4j. */
	protected void initArgs(String[] args)
	{
		CmdLineParser cmd = new CmdLineParser(this);
		
		try
		{
			cmd.parseArgument(args);
		}
		catch (CmdLineException e)
		{
			System.err.println(e.getMessage());
			cmd.printUsage(System.err);
			System.exit(1);
		}
		catch (Exception e) {e.printStackTrace();}
	}
	
	// ============================= getter: language =============================
	
	protected String getLanguage(Element eConfig)
	{
		Element eLanguage = UTXml.getFirstElementByTagName(eConfig, TAG_LANGUAGE);
		return UTXml.getTrimmedTextContent(eLanguage);
	}
	
	protected String getMode(Element eConfig)
	{
		Element eMode = UTXml.getFirstElementByTagName(eConfig, TAG_MODE);
		return UTXml.getTrimmedTextContent(eMode);
	}
	
	// ============================= getter: filenames =============================
	
	/** String[0]: input filename, String[1]: output filename. */
	protected List<String[]> getFilenames(String inputPath, String inputExt, String outputExt)
	{
		List<String[]> filenames = new ArrayList<String[]>();
		File f = new File(inputPath);
		String outputFile;
		
		if (f.isDirectory())
		{
			for (String inputFile : f.list(new FileExtFilter(inputExt)))
			{
				inputFile  = inputPath + File.separator + inputFile;
				outputFile = inputFile + "." + outputExt;
				filenames.add(new String[]{inputFile, outputFile});
			}
		}
		else
			filenames.add(new String[]{inputPath, inputPath+"."+outputExt});
		
		return filenames;
	}
	
	// ============================= getter: components =============================
	
	protected AbstractComponent getComponent(Element eModels, String mode) throws Exception
	{
		NodeList list = eModels.getElementsByTagName(TAG_MODEL);
		int i, size = list.getLength();
		Element eModel;
		
		for (i=0; i<size; i++)
		{
			eModel = (Element)list.item(i);
			
			if (mode.equals(UTXml.getTrimmedAttribute(eModel, TAG_TYPE)))
				return EngineGetter.getComponent(UTXml.getTrimmedAttribute(eModel, TAG_PATH), mode);
		}
		
		return null;
	}
	
	// ============================= getter: readers =============================
	
	protected Pair<AbstractReader<?>,String> getReader(Element eReader)
	{
		String type = UTXml.getTrimmedAttribute(eReader, TAG_TYPE);
		
		if      (type.equals(AbstractReader.TYPE_RAW))
			return new Pair<AbstractReader<?>,String>(new RawReader(), type);
		else if (type.equals(AbstractReader.TYPE_LINE))
			return new Pair<AbstractReader<?>,String>(new LineReader(), type);
		else
			return new Pair<AbstractReader<?>,String>(getJointReader(eReader), type);
	}
	
	protected JointReader getJointReader(Element eReader)
	{
		ObjectIntOpenHashMap<String> map = getFieldMap(eReader);
		
		int iId		= map.get(AbstractColumnReader.FIELD_ID)	 - 1;
		int iForm	= map.get(AbstractColumnReader.FIELD_FORM)	 - 1;
		int iLemma	= map.get(AbstractColumnReader.FIELD_LEMMA)	 - 1;
		int iPos	= map.get(AbstractColumnReader.FIELD_POS)	 - 1;
		int iFeats	= map.get(AbstractColumnReader.FIELD_FEATS)	 - 1;
		int iHeadId	= map.get(AbstractColumnReader.FIELD_HEADID) - 1;
		int iDeprel	= map.get(AbstractColumnReader.FIELD_DEPREL) - 1;
		int iSHeads = map.get(AbstractColumnReader.FIELD_SHEADS) - 1;
		int iNamex  = map.get(AbstractColumnReader.FIELD_NAMENT) - 1;
		
		return new JointReader(iId, iForm, iLemma, iPos, iFeats, iHeadId, iDeprel, iSHeads, iNamex);
	}
	
	/** Called by {@link AbstractBin#getCDEPReader(Element, String)}. */
	private ObjectIntOpenHashMap<String> getFieldMap(Element eReader)
	{
		NodeList list = eReader.getElementsByTagName(TAG_READER_COLUMN);
		int i, index, size = list.getLength();
		Element element;
		String field;
		
		ObjectIntOpenHashMap<String> map = new ObjectIntOpenHashMap<String>();
		
		for (i=0; i<size; i++)
		{
			element = (Element)list.item(i);
			field   = UTXml.getTrimmedAttribute(element, TAG_READER_COLUMN_FIELD);
			index   = Integer.parseInt(element.getAttribute(TAG_READER_COLUMN_INDEX));
			
			map.put(field, index);
		}
		
		return map;
	}
	
	// ============================= classification =============================
	
	protected StringTrainSpace mergeTrainSpaces(List<StringTrainSpace> spaces, int labelCutoff, int featureCutoff)
	{
		StringTrainSpace space;
		
		if (spaces.size() == 1)
		{
			space = spaces.get(0);
		}
		else
		{
			System.out.println("Merging training instances:");
			space = new StringTrainSpace(false, labelCutoff, featureCutoff);
			
			for (StringTrainSpace s : spaces)
			{
				space.appendSpace(s);
				System.out.print(".");
				s.clear();
			}
			
			System.out.println();			
		}
		
		return space;
	}
	
	protected AbstractModel getModel(Element eTrain, AbstractTrainSpace space, int index)
	{
		NodeList list = eTrain.getElementsByTagName(TAG_TRAIN_ALGORITHM);
		int numThreads = getNumOfThreads(eTrain);
		Element  eAlgorithm;
		String   name;
		
		eAlgorithm = (Element)list.item(index);
		name       = UTXml.getTrimmedAttribute(eAlgorithm, TAG_TRAIN_ALGORITHM_NAME);
		
		if (name.equals("liblinear"))
		{
			byte solver = Byte  .parseByte  (UTXml.getTrimmedAttribute(eAlgorithm, "solver"));
			double cost = Double.parseDouble(UTXml.getTrimmedAttribute(eAlgorithm, "cost"));
			double eps  = Double.parseDouble(UTXml.getTrimmedAttribute(eAlgorithm, "eps"));
			double bias = Double.parseDouble(UTXml.getTrimmedAttribute(eAlgorithm, "bias"));

			return getLiblinearModel(space, numThreads, solver, cost, eps, bias);
		}
		else if (name.equals("adagrad"))
		{
			int    iter  = Integer.parseInt   (UTXml.getTrimmedAttribute(eAlgorithm, "iter"));
			double alpha = Double .parseDouble(UTXml.getTrimmedAttribute(eAlgorithm, "alpha"));
			double delta = Double .parseDouble(UTXml.getTrimmedAttribute(eAlgorithm, "delta"));
			
			return getAdaGradModel(space, numThreads, iter, alpha, delta);
		}
		
		return null;
	}
	
	/** Called by {@link AbstractBin#getModel(Element, AbstractTrainSpace, int)}. */
	protected AbstractModel getLiblinearModel(AbstractTrainSpace space, int numThreads, byte solver, double cost, double eps, double bias)
	{
		space.build();
		System.out.println("Liblinear:");
		System.out.printf("- solver=%d, cost=%f, eps=%f, bias=%f\n", solver, cost, eps, bias);
		return LiblinearTrain.getModel(space, numThreads, solver, cost, eps, bias);
	}
	
	/** Called by {@link AbstractBin#getModel(Element, AbstractTrainSpace, int)}. */
	protected AbstractModel getAdaGradModel(AbstractTrainSpace space, int numThreads, int iter, double alpha, double delta)
	{
		space.build();
		System.out.println("AdaGrad:");
		System.out.printf("- iter=%d, alpha=%f, delta=%f\n", iter, alpha, delta);
		
		AdaGrad ag = new AdaGrad(iter, alpha, delta);
		AbstractModel model = space.getModel();
		
		System.out.println("Training:");
		model.setWeights(ag.getWeight(space, numThreads));
		
		return model;
	}
	
	protected void updateModel(Element eTrain, AbstractTrainSpace space, IntObjectOpenHashMap<IntDoubleMap> mMargin, int nUpdate, int index)
	{
		NodeList list = eTrain.getElementsByTagName(TAG_TRAIN_ALGORITHM);
		int numThreads = getNumOfThreads(eTrain);
		Element  eAlgorithm;
		String   name;
		
		eAlgorithm = (Element)list.item(index);
		name       = UTXml.getTrimmedAttribute(eAlgorithm, TAG_TRAIN_ALGORITHM_NAME);
		
		if (name.equals("adagrad"))
		{
			int    iter  = Integer.parseInt   (UTXml.getTrimmedAttribute(eAlgorithm, "iter"));
			double alpha = Double .parseDouble(UTXml.getTrimmedAttribute(eAlgorithm, "alpha"));
			double delta = Double .parseDouble(UTXml.getTrimmedAttribute(eAlgorithm, "delta"));
			
			updateAdaGradModel(space, numThreads, mMargin, nUpdate, iter, alpha, delta);
		}
	}
	
	protected void updateAdaGradModel(AbstractTrainSpace space, int numThreads, IntObjectOpenHashMap<IntDoubleMap> mMargin, int nUpdate, int iter, double alpha, double delta)
	{
		AbstractModel model = space.getModel();
		
		if (model.getWeights() == null)
		{
			space.build();
			model.initWeightVector();
		}
		
		System.out.printf("%3d: AdaGrad, iter=%d, alpha=%f, delta=%f\n", nUpdate, iter, alpha, delta);
		AdaGrad ag = new AdaGrad(iter, alpha, delta);
		ag.updateWeight(space);
	}
	
	protected int getNumOfThreads(Element eTrain)
	{
		Element eThreads = UTXml.getFirstElementByTagName(eTrain, TAG_TRAIN_THREADS); 
		return Integer.parseInt(UTXml.getTrimmedTextContent(eThreads));
	}
	
	protected void printTime(String message, long st, long et)
	{
		long millis = et - st;
		long mins   = TimeUnit.MILLISECONDS.toMinutes(millis);
		long secs   = TimeUnit.MILLISECONDS.toSeconds(millis) - TimeUnit.MINUTES.toSeconds(mins);

		System.out.println(message);
		System.out.println(String.format("- %d mins, %d secs", mins, secs));
	}
}